
from __future__ import unicode_literals

from collections import OrderedDict
import json
import logging
import sys
import re

try:
    # pylint: disable=no-name-in-module
    from urllib.parse import urlencode
except ImportError:
    # pylint: disable=no-name-in-module
    from urllib import urlencode

from .exceptions import Error, ProgrammingError

from .row import Row
from .extensions import _convert_to_python, _adapt_from_python, _column_stripper


if sys.version_info[0] >= 3:
    basestring = str
    _urlencode = urlencode
else:
    # avoid UnicodeEncodeError from urlencode
    def _urlencode(query, doseq=0):
        return urlencode(dict(
            (k if isinstance(k, bytes) else k.encode('utf-8'),
             v if isinstance(v, bytes) else v.encode('utf-8'))
            for k, v in query.items()), doseq=doseq)


class Cursor(object):
    arraysize = 1

    def __init__(self, connection):
        self._connection = connection
        self.messages = []
        self.lastrowid = None
        self.description = None
        self.rownumber = 0
        self.rowcount = -1
        self.arraysize = 1
        self._rows = None
        self._column_type_cache = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    @property
    def connection(self):
        return self._connection

    def close(self):
        self._rows = None

    def _request(self, method, uri, body=None, headers={}):
        logger = logging.getLogger(__name__)
        debug = logger.getEffectiveLevel() < logging.DEBUG
        logger.debug(
            'request method: %s uri: %s headers: %s body: %s',
            method,
            uri,
            headers,
            body)
        response = self.connection._fetch_response(
            method, uri, body=body, headers=headers)
        logger.debug(
            "status: %s reason: %s",
            response.status,
            response.reason)
        response_text = response.read().decode('utf-8')
        logger.debug("raw response: %s", response_text)
        response_json = json.loads(
            response_text, object_pairs_hook=OrderedDict)
        if debug:
            logger.debug(
                "formatted response: %s",
                json.dumps(
                    response_json,
                    indent=4))
        return response_json

    def _substitute_params(self, operation, parameters):
        '''
        SQLite natively supports only the types TEXT, INTEGER, REAL, BLOB and
        NULL
        '''

        param_matches = 0

        qmark_re = re.compile(r"(\?)")
        named_re = re.compile(r"(:{1}[a-zA-Z]+?\b)")

        qmark_matches = qmark_re.findall(operation)
        named_matches = named_re.findall(operation)
        param_matches = len(qmark_matches) + len(named_matches)

        # Matches but no parameters
        if param_matches > 0 and parameters is None:
            raise ProgrammingError('paramater required but not given: %s' %
                                   operation)

        # No regex matches and no parameters.
        if parameters is None:
            return operation

        if len(qmark_matches) > 0 and len(named_matches) > 0:
            raise ProgrammingError('different paramater types in operation not'
                                   'permitted: %s %s' % 
                                   (operation, parameters))

        if isinstance(parameters, dict):
            # parameters is a dict or a dict subclass
            if len(qmark_matches) > 0:
                raise ProgrammingError('Unamed binding used, but you supplied '
                                       'a dictionary (which has only names): '
                                       '%s %s' % (operation, parameters))
            for op_key in named_matches:
                try:
                    operation = operation.replace(op_key, 
                                                 _adapt_from_python(parameters[op_key[1:]]))
                except KeyError:
                    raise ProgrammingError('the named parameters given do not '
                                           'match operation: %s %s' %
                                           (operation, parameters))
        else:
            # parameters is a sequence
            if param_matches != len(parameters):
                raise ProgrammingError('incorrect number of parameters '
                                       '(%s != %s): %s %s' % (param_matches, 
                                       len(parameters), operation, parameters))
            if len(named_matches) > 0:
                raise ProgrammingError('Named binding used, but you supplied a'
                                       ' sequence (which has no names): %s %s' %
                                       (operation, parameters))
            parts = operation.split('?')
            subst = []
            for i, part in enumerate(parts):
                subst.append(part)
                if i < len(parameters):
                    subst.append(_adapt_from_python(parameters[i]))
            operation = ''.join(subst)

        return operation

    def _get_sql_command(self, sql_str):
        return sql_str.split(None, 1)[0].upper()

    def execute(self, operation, parameters=None):
        if not isinstance(operation, basestring):
            raise ValueError(
                             "argument must be a string, not '{}'".format(type(operation).__name__))

        operation = self._substitute_params(operation, parameters)

        command = self._get_sql_command(operation)
        if command in ('SELECT', 'PRAGMA'):
            payload = self._request("GET",
                                    "/db/query?" + _urlencode({'q': operation}))
        else:
            payload = self._request("POST", "/db/execute?transaction",
                                    headers={'Content-Type': 'application/json'}, body=json.dumps([operation]))

        last_insert_id = None
        rows_affected = -1
        payload_rows = {}
        try:
            results = payload["results"]
        except KeyError:
            pass
        else:
            rows_affected = 0
            for item in results:
                if 'error' in item:
                    logging.getLogger(__name__).error(json.dumps(item))
                    raise Error(json.dumps(item))
                try:
                    rows_affected += item['rows_affected']
                except KeyError:
                    pass
                try:
                    last_insert_id = item['last_insert_id']
                except KeyError:
                    pass
                if 'columns' in item:
                    payload_rows = item

        try:
            fields = payload_rows['columns']
        except KeyError:
            self.description = None
            self._rows = []
            if command == 'INSERT':
                self.lastrowid = last_insert_id
        else:
            rows = []
            description = []
            for field in fields:
                description.append((
                    _column_stripper(field, parse_colnames=self.connection.parse_colnames),
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ))

            try:
                values = payload_rows['values']
                types = payload_rows['types']
            except KeyError:
                pass
            else:
                if values:
                    converters = [_convert_to_python(field, type_,
                                  parse_decltypes=self.connection.parse_decltypes,
                                  parse_colnames=self.connection.parse_colnames)
                                  for field, type_ in zip(fields, types)]
                    for payload_row in values:
                        row = []
                        for field, converter, value in zip(fields, converters, payload_row):
                            row.append((field, (value if converter is None
                                                else converter(value))))
                        rows.append(Row(row))
            self._rows = rows
            self.description = tuple(description)

        self.rownumber = 0
        if command in ('UPDATE', 'DELETE'):
            # sqalchemy's _emit_update_statements function asserts
            # rowcount for each update, and _emit_delete_statements
            # warns unless rowcount matches
            self.rowcount = rows_affected
        else:
            self.rowcount = len(self._rows)
        return self

    def executemany(self, operation, seq_of_parameters=None):
        if not isinstance(operation, basestring):
            raise ValueError(
                "argument must be a string, not '{}'".format(type(operation).__name__))

        statements = []
        for parameters in seq_of_parameters:
            statements.append(self._substitute_params(operation, parameters))
        payload = self._request("POST", "/db/execute?transaction",
                                headers={'Content-Type': 'application/json'},
                                body=json.dumps(statements))
        rows_affected = -1
        try:
            results = payload["results"]
        except KeyError:
            pass
        else:
            rows_affected = 0
            for item in results:
                if 'error' in item:
                    logging.getLogger(__name__).error(json.dumps(item))
                try:
                    rows_affected += item['rows_affected']
                except KeyError:
                    pass
        self._rows = []
        self.rownumber = 0
        self.rowcount = rows_affected

    def fetchone(self):
        ''' Fetch the next row '''
        if self._rows is None or self.rownumber >= len(self._rows):
            return None
        result = self._rows[self.rownumber]
        self.rownumber += 1
        return result

    def fetchmany(self, size=None):
        remaining = self.arraysize if size is None else size
        remaining = min(remaining, self.rowcount - self.rownumber)
        return [self.fetchone() for i in range(remaining)]

    def fetchall(self):
        rows = []
        while self.rownumber < self.rowcount:
            rows.append(self.fetchone())
        return rows

    def setinputsizes(self, sizes):
        raise NotImplementedError(self)

    def setoutputsize(self, size, column=None):
        raise NotImplementedError(self)

    def scroll(self, value, mode='relative'):
        raise NotImplementedError(self)

    def next(self):
        raise NotImplementedError(self)

    def __iter__(self):
        while self.rownumber < self.rowcount:
            yield self.fetchone()




try:
    # pylint: disable=no-name-in-module
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from collections import OrderedDict


class Row(tuple, Mapping):

    def __new__(cls, items):
        return tuple.__new__(cls, (item[1] for item in items))

    def __init__(self, items):
        super(Row, self).__init__()
        self._dict = OrderedDict(items)

    def __getitem__(self, k):
        try:
            return self._dict[k]
        except (KeyError, TypeError):
            if isinstance(k, (int, slice)):
                return tuple.__getitem__(self, k)
            else:
                raise

    def __iter__(self):
        return tuple.__iter__(self)

    def keys(self):
        for k in self._dict:
            if not isinstance(k, int):
                yield k

    def __len__(self):
        return tuple.__len__(self)

    def __str__(self):
        return str(self._dict)

    def __delitem__(self, k):
        raise NotImplementedError(self)

    def pop(self, k):
        raise NotImplementedError(self)




from __future__ import unicode_literals

import codecs
import logging

try:
    from http.client import HTTPConnection, HTTPSConnection
except ImportError:
    # pylint: disable=import-error
    from httplib import HTTPConnection, HTTPSConnection

try:
    from urllib.parse import urlparse
except ImportError:
    # pylint: disable=import-error
    from urlparse import urlparse

from .constants import (
    UNLIMITED_REDIRECTS,
)

from .cursors import Cursor
from ._ephemeral import EphemeralRqlited as _EphemeralRqlited
from .extensions import PARSE_DECLTYPES, PARSE_COLNAMES


class Connection(object):

    from .exceptions import (
        Warning,
        Error,
        InterfaceError,
        DatabaseError,
        DataError,
        OperationalError,
        IntegrityError,
        InternalError,
        ProgrammingError,
        NotSupportedError,
    )

    def __init__(self, scheme='http', host='localhost', port=4001,
                 user=None, password=None, connect_timeout=None,
                 detect_types=0, max_redirects=UNLIMITED_REDIRECTS):

        self.messages = []
        self.scheme = scheme
        self.host = host
        self.port = port
        self._headers = {}
        if not (user is None or password is None):
            self._headers['Authorization'] = 'Basic ' + \
                codecs.encode('{}:{}'.format(user, password).encode('utf-8'),
                              'base64').decode('utf-8').rstrip('\n')

        self.connect_timeout = connect_timeout
        self.max_redirects = max_redirects
        self.detect_types = detect_types
        self.parse_decltypes = detect_types & PARSE_DECLTYPES
        self.parse_colnames = detect_types & PARSE_COLNAMES
        self._ephemeral = None
        if scheme == ':memory:':
            self._ephemeral = _EphemeralRqlited().__enter__()
            self.host, self.port = self._ephemeral.http
        self._connection = self._init_connection()

    def _init_connection(self):
        if self.scheme in ('http', ':memory:'):
            cls = HTTPConnection
        elif self.scheme == 'https':
            cls = HTTPSConnection
        else:
            raise Connection.ProgrammingError('Unsupported scheme %r' % self.scheme)
        return cls(self.host, port=self.port,
                   timeout=None if self.connect_timeout is None else float(self.connect_timeout))

    def _retry_request(self, method, uri, body=None, headers={}):
        tries = 10
        while tries:
            tries -= 1
            try:
                self._connection.request(method, uri, body=body,
                                         headers=dict(self._headers, **headers))
                return self._connection.getresponse()
            except Exception:
                if not tries:
                    raise
                self._connection.close()
                self._connection = self._init_connection()

    def _fetch_response(self, method, uri, body=None, headers={}):
        """
        Fetch a response, handling redirection.
        """
        response = self._retry_request(method, uri, body=body, headers=headers)
        redirects = 0

        while response.status == 301 and \
                response.getheader('Location') is not None and \
                (self.max_redirects == UNLIMITED_REDIRECTS or redirects < self.max_redirects):
            redirects += 1
            uri = response.getheader('Location')
            location = urlparse(uri)

            logging.getLogger(__name__).debug("status: %s reason: '%s' location: '%s'",
                                              response.status, response.reason, uri)

            if self.host != location.hostname or self.port != location.port:
                self._connection.close()
                self.host = location.hostname
                self.port = location.port
                self._connection = self._init_connection()

            response = self._retry_request(method, uri, body=body, headers=headers)

        return response

    def close(self):
        """Close the connection now (rather than whenever .__del__() is
        called).

        The connection will be unusable from this point forward; an
        Error (or subclass) exception will be raised if any operation
        is attempted with the connection. The same applies to all
        cursor objects trying to use the connection. Note that closing
        a connection without committing the changes first will cause an
        implicit rollback to be performed."""
        self._connection.close()
        if self._ephemeral is not None:
            self._ephemeral.__exit__(None, None, None)
            self._ephemeral = None

    def __del__(self):
        self.close()

    def commit(self):
        """Database modules that do not support transactions should
        implement this method with void functionality."""
        pass

    def rollback(self):
        """This method is optional since not all databases provide
        transaction support. """
        pass

    def cursor(self, factory=None):
        """Return a new Cursor Object using the connection."""
        if factory:
            return factory(self)
        else:
            return Cursor(self)

    def execute(self, *args, **kwargs):
        return self.cursor().execute(*args, **kwargs)




import contextlib
import errno
import os
import shutil
import socket
import subprocess
import sys
import tempfile
import time


class EphemeralRqlited(object):
    def __init__(self):
        self.host = None
        self.http = None
        self.raft = None
        self._tempdir = None
        self._proc = None

    @staticmethod
    def _unused_ports(host, count):
        sockets = []
        ports = []
        try:
            sockets.extend(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                for i in range(count))
            for s in sockets:
                s.bind((host, 0))
                ports.append(s.getsockname()[-1])
        finally:
            while sockets:
                sockets.pop().close()

        return ports

    @staticmethod
    def _test_port(host, port, timeout=None):
        try:
            with contextlib.closing(
                socket.create_connection((host, port), timeout=timeout)):
                return True
        except socket.error:
            return False

    def _start(self):
        self._tempdir = tempfile.mkdtemp()
        self.host = 'localhost'

        # Allocation of unused ports is racy, so retry
        # until ports have been successfully acquired.
        while self._proc is None:
            http_port, raft_port = self._unused_ports(self.host, 2)
            self.http = (self.host, http_port)
            self.raft = (self.host, raft_port)
            with open(os.devnull, mode='wb', buffering=0) as devnull:
                filename = 'rqlited'
                try:
                    self._proc = subprocess.Popen([filename,
                        '-http-addr', '{}:{}'.format(*self.http),
                        '-raft-addr', '{}:{}'.format(*self.raft), self._tempdir],
                        stdout=devnull, stderr=devnull)
                except EnvironmentError as e:
                    if e.errno == errno.ENOENT and sys.version_info.major < 3:
                        # Add filename to clarify exception message.
                        e.filename = filename
                    raise

            while not self._test_port(*self.http) and self._proc.poll() is None:
                time.sleep(0.5)

            if self._proc.poll() is not None:
                self._proc = None

    def __enter__(self):
        self._start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._tempdir is not None:
            shutil.rmtree(self._tempdir)
            self._tempdir = None
        if self._proc is not None:
            self._proc.terminate()
            self._proc.wait()
            self._proc = None
        return False




from __future__ import unicode_literals

import time

from .constants import (
    UNLIMITED_REDIRECTS,
)

from .connections import Connection
connect = Connection

from .exceptions import (
    Warning,
    Error,
    InterfaceError,
    DataError,
    DatabaseError,
    OperationalError,
    IntegrityError,
    InternalError,
    NotSupportedError,
    ProgrammingError,
)

from .types import (
    Binary,
    Date,
    Time,
    Timestamp,
    STRING,
    BINARY,
    NUMBER,
    DATETIME,
    ROWID,
)

# Compat with native sqlite module
from .extensions import converters, adapters, register_converter, register_adapter
from sqlite3.dbapi2 import PrepareProtocol


paramstyle = "qmark"

threadsafety = 1

apilevel = "2.0"


def DateFromTicks(ticks):
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return Timestamp(*time.localtime(ticks)[:6])

# accessed by sqlalchemy sqlite dialect
sqlite_version_info = (3, 10, 0)

# Compat with native sqlite module
from .extensions import PARSE_DECLTYPES, PARSE_COLNAMES



#!/usr/bin/env python

import os
from os.path import isdir, islink, relpath, dirname
import subprocess
import sys
from setuptools import (
    Command,
    setup,
    find_packages,
)

sys.path.insert(0, 'src')
from pyrqlite.constants import (
    __author__,
    __email__,
    __license__,
    __version__,
)

class PyTest(Command):
    user_options = [('match=', 'k', 'Run only tests that match the provided expressions')]

    def initialize_options(self):
        self.match = None

    def finalize_options(self):
        pass

    def run(self):
        testpath = 'src/test'
        buildlink = 'build/lib/test'

        if isdir(dirname(buildlink)):
            if islink(buildlink):
                os.unlink(buildlink)

            os.symlink(relpath(testpath, dirname(buildlink)), buildlink)
            testpath = buildlink

        try:
            os.environ['EPYTHON'] = 'python{}.{}'.format(sys.version_info.major, sys.version_info.minor)
            subprocess.check_call(['py.test', '-v', testpath, '-s',
                        '--cov-report=html', '--cov-report=term-missing'] +
                       (['-k', self.match] if self.match else []) +
                       ['--cov={}'.format(p) for p in find_packages(dirname(testpath), exclude=['test'])])

        finally:
            if islink(buildlink):
                os.unlink(buildlink)


class PyLint(Command):
    user_options = [('errorsonly', 'E', 'Check only errors with pylint'),
                    ('format=', 'f', 'Change the output format')]

    def initialize_options(self):
        self.errorsonly = 0
        self.format = 'colorized'

    def finalize_options(self):
        pass

    def run(self):
        cli_options = ['-E'] if self.errorsonly else []
        cli_options.append('--output-format={0}'.format(self.format))
        errno = subprocess.call(['pylint'] + cli_options + [
            "--msg-template='{C}:{msg_id}:{path}:{line:3d},{column}: {obj}: {msg} ({symbol})'"] +
            find_packages('src', exclude=['test']), cwd='./src')
        raise SystemExit(errno)


setup(
    name="pyrqlite",
    version=__version__,
    url='https://github.com/rqlite/pyrqlite/',
    author=__author__,
    author_email=__email__,
    maintainer=__author__,
    maintainer_email=__email__,
    description='python DB API 2.0 driver for rqlite',
    license=__license__,
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['test']),
    platforms=['Posix'],
    cmdclass={'test': PyTest, 'lint': PyLint},
    tests_require=['pytest', 'pytest-cov'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Database',
    ],
)




from __future__ import unicode_literals

__version__ = "2.1"
__project__ = "pyrqlite"

__author__ = "Zac Medico"
__email__ = "zmedico@gmail.com"

__copyright__ = "Copyright (C) 2016 Zac Medico"
__license__ = "MIT"
__description__ = "Python dbapi2 driver for rqlite"

UNLIMITED_REDIRECTS = -1
